import { log } from 'console';
import { SchemaAttribute, SchemaGraph, SchemaGraphologyEdge, SchemaGraphologyNode } from '../../../schema';
import { AttributeNames, EntitiesFromSchema, RelationsFromSchema } from '../Types';

/**
 * Takes a schema result and calculates all the entity names, and relation names and attribute names per entity.
 * Used by PAOHvis to show all possible options to choose from when adding a new PAOHvis visualisation or when filtering.
 * @param {SchemaResultType} schemaResult A new schema result from the backend.
 * @returns {EntitiesFromSchema} All entity names, and relation names and attribute names per entity.
 */
export function calculateAttributesAndRelations(schemaResult: SchemaGraph): EntitiesFromSchema {
  const attributesPerEntity: Record<string, AttributeNames> = calculateAttributes(schemaResult);
  const relationsPerEntity: Record<string, string[]> = calculateRelations(schemaResult);

  return {
    entityNames: schemaResult.nodes.map((node) => node?.attributes?.name || 'ERROR'),
    attributesPerEntity,
    relationsPerEntity,
  };
}

/**
 * Takes a schema result and calculates all the attribute names per entity.
 * @param {SchemaResultType} schemaResult A new schema result from the backend.
 * @returns {Record<string, AttributeNames>} All attribute names per entity.
 */
export function calculateAttributes(schemaResult: SchemaGraph): Record<string, AttributeNames> {
  const attributesPerEntity: Record<string, AttributeNames> = {};
  // Go through each entity.
  schemaResult.nodes.forEach((node) => {
    if (node?.attributes?.name === undefined) {
      console.error('ERROR: Node has no name attribute or name.', node);
      return;
    }

    // Extract the attribute names per datatype for each entity.
    const textAttributeNames: string[] = [];
    const boolAttributeNames: string[] = [];
    const numberAttributeNames: string[] = [];

    node.attributes.attributes.forEach((attr) => {
      if (attr.type == 'string') textAttributeNames.push(attr.name);
      else if (attr.type == 'int' || attr.type == 'float') numberAttributeNames.push(attr.name);
      else boolAttributeNames.push(attr.name);
    });

    // Create a new object with the arrays with attribute names per datatype.
    attributesPerEntity[node.attributes.name] = {
      textAttributeNames,
      boolAttributeNames,
      numberAttributeNames,
    };
  });
  return attributesPerEntity;
}

/**
 * Takes a schema result and calculates all the relation names per entity.
 * @param {SchemaResultType} schemaResult A new schema result from the backend.
 * @returns {Record<string, AttributeNames>} All relation (from and to) names per entity.
 */
export function calculateRelations(schemaResult: SchemaGraph): Record<string, string[]> {
  const relationsPerEntity: Record<string, string[]> = {};
  // Go through each relation.
  schemaResult.edges.forEach((edge) => {
    if (edge?.attributes === undefined) {
      console.error('ERROR: Edge has no attribute.', edge);
      return;
    }

    // Extract the from-node-name (collection name) from every relation.
    if (relationsPerEntity[edge.attributes.from]) relationsPerEntity[edge.attributes.from].push(edge.attributes.collection);
    else relationsPerEntity[edge.attributes.from] = [edge.attributes.collection];
    // Extract the to-node-name (collection name) from every relation.
    if (relationsPerEntity[edge.attributes.to]) relationsPerEntity[edge.attributes.to].push(edge.attributes.collection);
    else relationsPerEntity[edge.attributes.to] = [edge.attributes.collection];
  });
  return relationsPerEntity;
}

/**
 * Takes a schema result and calculates all the relation collection names, and relation names and attribute names per relation.
 * Used by PAOHvis to show all possible options to choose from when adding a new PAOHvis visualisation or when filtering.
 * @param {SchemaResultType} schemaResult A new schema result from the backend.
 * @returns {EntitiesFromSchema} All entity names, and relation names and attribute names per entity.
 */
export function calculateAttributesFromRelation(schemaResult: SchemaGraph): RelationsFromSchema {
  const relationCollections: string[] = [];
  const attributesPerRelation: Record<string, AttributeNames> = {};
  const nameOfCollectionPerRelation: Record<string, string> = {};
  // Go through each relation.
  schemaResult.edges.forEach((edge) => {
    if (edge?.attributes === undefined) {
      console.error('ERROR: Edge has no attribute.', edge);
      return;
    }

    if (!nameOfCollectionPerRelation[edge.attributes.collection]) {
      relationCollections.push(edge.attributes.collection);
      nameOfCollectionPerRelation[edge.attributes.collection] = `${edge.attributes.from}:${edge.attributes.to}:${edge.attributes.name}`;
    }
    // Extract the attribute names per datatype for each relation.
    const textAttributeNames: string[] = [];
    const boolAttributeNames: string[] = [];
    const numberAttributeNames: string[] = [];
    edge.attributes.attributes.forEach((attr: SchemaAttribute) => {
      if (attr.type == 'string') textAttributeNames.push(attr.name);
      else if (attr.type == 'int' || attr.type == 'float') numberAttributeNames.push(attr.name);
      else boolAttributeNames.push(attr.name);
    });

    // Create a new object with the arrays with attribute names per datatype.
    attributesPerRelation[edge.attributes.collection] = {
      textAttributeNames,
      boolAttributeNames,
      numberAttributeNames,
    };
  });
  return {
    relationCollection: relationCollections,
    relationNames: nameOfCollectionPerRelation,
    attributesPerRelation: attributesPerRelation,
  };
}

/**
 * Filters duplicate elements from array with a hashtable.
 * From https://stackoverflow.com/questions/9229645/remove-duplicate-values-from-js-array */
export function uniq(element: number[]) {
  const seen: Record<number, boolean> = {};
  return element.filter(function (item) {
    return seen.hasOwnProperty(item) ? false : (seen[item] = true);
  });
}

/**
 * Calculate the width of the specified text.
 * @param txt Text input as string.
 * @param fontname Name of the font.
 * @param fontsize Size of the fond in px.
 * @param fontWeight The weight of the font.
 * @returns {number} Width of the textfield in px.
 */
export const getWidthOfText = (txt: string, fontname: string, fontsize: string, fontWeight = 'normal'): number => {
  let c = document.createElement('canvas');
  let ctx = c.getContext('2d') as CanvasRenderingContext2D;
  let fontspec = fontWeight + ' ' + fontsize + ' ' + fontname;
  if (ctx.font !== fontspec) ctx.font = fontspec;
  return ctx.measureText(txt).width;
};
